#ifndef HEAP_HPP_
#define HEAP_HPP_
#define heap_hpp_version "$Revision: 1.8 $"

#include "Assertion.hpp"
#include "BitVector.hpp"
#include	"Random.hpp"
#include <vector>
#include "EInfo.hpp"
#include "Logging.hpp"



template<class ValueType>
struct ValCompPrefMin{
	bool better(const ValueType& a, const ValueType& b)const{return a<b;}
};

template<class ValueType>
struct ValCompPrefMax{
	bool better(const ValueType& a, const ValueType& b)const{return a>b;}
};


/*template<class KeyType>
struct KeyTypeConverter{
	this function is required to store elements of type KeyType in the heap:
	uint32_t toInt(const KeyType& k)const;

	this function is only required by the function insertElementsTo
	KeyType fromInt(const uint32_t& i)const;
};*/




/**
 * Heap structure that ranks all elements according to some value.*/
template<class KeyType, class ValueType, class ValueCompare = ValCompPrefMin<ValueType>, class KeyTypeConverter = DfltKeyTypeConverter<KeyType> >
class Heap{
	static inline uint32_t   parent		(uint32_t index)	 	{return (index-1) >> 1;};
	static inline uint32_t   leftChild	 (uint32_t index)		{return (index << 1) + 1;};
	static inline uint32_t   rightChild	(uint32_t index)	 	{return (index << 1) + 2;};
	static const uint32_t  	IVLD_IDX = ~0;

protected:
	const KeyTypeConverter& 	converter;
	const ValueCompare&			comp;

	/**Internal data-structure to store the information for one element.*/
	struct ElementInfo{
		uint32_t 	index_in_vec;
		ValueType 	value;

		/**Constructs an invalid emement.*/
		inline ElementInfo() :index_in_vec(IVLD_IDX),value(0){};
		inline ElementInfo(uint32_t idx,const ValueType& v):index_in_vec(idx),value(v){}
		inline bool contained()const {return index_in_vec != IVLD_IDX;}
	};


	/**Holds an ElementInfo for each element in the heap.*/
	std::vector<ElementInfo> 	e_info;

	inline uint32_t	moveUp(const uint32_t element_id, uint32_t heap_index);
	inline uint32_t	moveDown(const uint32_t element_id, uint32_t heap_index);
	inline void		updateAtPos(uint32_t pos);

protected:
	/**The actual heap. Each element of the heap vector is an index of the original element into e_info.*/
	std::vector<KeyType> 			eles;

	inline			Heap(const Heap& h);

	/**No assignment.*/
	inline const	Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter>& operator=(const Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter>& h);

	/**Returns element id of the element that was at the specified position pos.
	 * Does not ensure the heap property*/
	inline KeyType	lazyRemoveAtPos(uint32_t pos);

	inline bool		checkHeapProperty(uint32_t heap_index) const;

	template<class PV>
	inline void		printHeap(PV& val_printer, uint32_t heap_index, uint32_t levl,uint32_t parnt) const;

	inline bool		better(uint32_t el_a, uint32_t el_b) const
					{return comp.better(e_info[el_a].value,e_info[el_b].value);};

	/*inline bool		betterOrEq(uint32_t el_a, uint32_t el_b) const
					{return !comp.better(e_info[el_b].value,e_info[el_a].value);};*/


public:

	Heap(const ValueCompare& compare=ValueCompare(),const KeyTypeConverter& conv = KeyTypeConverter())
					:converter(conv),comp(compare){}

	/**Reserve enough space for the specified number of elements*/
	void			reserve(uint32_t capacity){
					if(capacity >= e_info.size()){
						e_info	.reserve(capacity);
						eles	.reserve(capacity);}}


	/**Validates the given element on the heap.
	 * If not yet contained, the initial value is set to default_val*/
	inline void		insertElementsTo(const uint32_t& req_sz, const ValueType& default_val);

	/**Ensures to have values for all elements up to max_element. New elements are not inserted into the heap*/
	inline void		ensureElementValsTo(const uint32_t& req_sz, const ValueType& default_val);

	/**Reactivates an element that was previously removed by the function removeTop or removeAtPos.*/
	inline void		restoreElementIf(const KeyType& element);

	/**Return the current rank value of the specified element*/
	ValueType 		getValue(const KeyType& element) const{
					asrt(converter.toInt(element) < e_info.size()); return e_info[converter.toInt(element)].value;}

	ValueType		getTop() const{
					return e_info[converter.toInt(eles[0])].value;}

	ValueType		getBest() const{
					return eles.empty() ? ValueType(0) : getTop();}

	void			setValue(const KeyType& e, const ValueType& v){
					asrt(converter.toInt(e) < e_info.size()); sanity(checkHeapProperty(0));
					e_info[converter.toInt(e)].value = v;
					if(e_info[converter.toInt(e)].contained()){
						updateAtPos(e_info[converter.toInt(e)].index_in_vec);}
						sanity(checkHeapProperty(0));}


	/**Checks if the heap is empty*/
	bool		 	empty()	{return eles.empty();}
	uint32_t 		size()	{return eles.size();}

	void			clear(){
					for(uint32_t i=eles.size();i--;){e_info[converter.toInt(eles[i])].index_in_vec=IVLD_IDX;}
					eles.clear();}

	bool			contains(const KeyType& e)const{
					asrt(converter.toInt(e) < e_info.size());
					return e_info[converter.toInt(e)].contained();}

	KeyType			getAtPos(uint32_t pos){
					asrt(pos < eles.size());
					return converter.fromInt(eles[pos]);}

	KeyType			removeAtPos(uint32_t pos){
					asrt(pos < eles.size());
					KeyType ret =	lazyRemoveAtPos(pos);
					if(pos < eles.size()){	updateAtPos(pos);}	sanity(checkHeapProperty(0));
					return ret;}

	void			removeIfContained(const KeyType& e){
					asrt(converter.toInt(e) < e_info.size());
					ElementInfo& ef = e_info[converter.toInt(e)];
					if(!ef.contained()) return;
					removeAtPos(ef.index_in_vec);sanity(checkHeapProperty(0));}

	/**Get the element with the highest rank value and remove it from the heap.*/
	KeyType			removeTop(){
					asrt(!empty());
					KeyType ret = lazyRemoveAtPos(0);
					if(eles.size() > 1){moveDown(converter.toInt(eles[0]),0);}sanity(checkHeapProperty(0));
					return ret;}

	KeyType			removeRandom(Random& randy){
					return 	removeAtPos(randy.nextInt(eles.size()));
	}

	bool			asrtHeapProperty() const {return checkHeapProperty(0);}
	template<class PV>
	void			printHeap(PV& val_printer) const {printHeap(val_printer, 0,0,IVLD_IDX);}
};



template<class KeyType, class ValueType, class ValueCompare,  class KeyTypeConverter = DfltKeyTypeConverter<KeyType> >
class BumpHeap: public Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter >{
	typedef Heap<KeyType,ValueType,ValueCompare,KeyTypeConverter > super;
	const ValueType	max;
	ValueType	   	increment;
	const ValueType decay;
	const typename ValueType::denominator_t weaken;
public:
	inline bool		activityBump(const KeyType& element);
	void			decayAll()				{increment *= decay; };
	BumpHeap(ValueType MAX, ValueType dec,const ValueCompare& comp):
		super(comp,KeyTypeConverter()),
		max(MAX),
		increment(1, ValueType::minNonZeroNumber().getLogOfDenominator() - (2* dec.getLogOfDenominator())),
		decay(dec),
		weaken(increment.getLogOfDenominator()){}
};



//Inline Implementation


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
uint32_t Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::moveUp(const uint32_t element_id, uint32_t heap_index){
	asrt(heap_index < eles.size()); asrt(e_info[element_id].index_in_vec == heap_index);
	const KeyType element = eles[heap_index];

	while( heap_index != 0){
		uint32_t parent_idx = parent(heap_index);
		uint32_t parent_id 	= converter.toInt(eles[parent_idx]);

		//if(!better(element_id,parent_id)){break;}
		if(better(parent_id,element_id)){break;}

		eles[heap_index]								=   eles[parent_idx];
		e_info[parent_id].index_in_vec					=   heap_index;
		asrt(e_info[parent_id].contained());
		heap_index										=   parent_idx;
	}
	eles[heap_index]									=   element;
	e_info[element_id].index_in_vec						=   heap_index;
	asrt(e_info[element_id].contained());
	return heap_index;
}

template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
uint32_t Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::moveDown(const uint32_t element_id, uint32_t heap_index){
	asrt(heap_index < eles.size()); asrt(e_info[element_id].index_in_vec == heap_index);
	const KeyType element = eles[heap_index];

	uint32_t child_index,child_id,right_id;
	while((child_index = leftChild(heap_index), child_index < eles.size())){

		child_id = converter.toInt(eles[child_index]);
		//get the child with the higher rank
		if(child_index+1 < eles.size() && !better(child_id,right_id = converter.toInt(eles[child_index+1]))){
			++child_index;
			child_id = right_id;
		} //the right one is bigger


		if(better(element_id,child_id))	break;

		eles[heap_index]								=   eles[child_index];
		e_info[child_id].index_in_vec					=   heap_index;
		asrt(e_info[child_id].contained());
		heap_index										=   child_index;
	}
	eles[heap_index]									=   element;
	e_info[element_id].index_in_vec						=   heap_index;
	return heap_index;
}


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::updateAtPos(uint32_t pos){
	asrt(pos < eles.size());
	const uint32_t eid = converter.toInt(eles[pos]);
	pos = moveDown(eid,pos); asrt(e_info[eid].index_in_vec==pos);
	moveUp(eid,pos);	asrt(e_info[eid].index_in_vec==pos);
	sanity(checkHeapProperty(0));
}



//has to be ensured that not empty
template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
KeyType Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::lazyRemoveAtPos(uint32_t pos){
	asrt(pos < eles.size()); asrt(!eles.empty());

	KeyType element					= eles[pos];
	KeyType last					= eles.back(); eles.pop_back();
	eles[pos]						= last;
	e_info[converter.toInt(last)]	.index_in_vec		= pos;
	e_info[converter.toInt(element)].index_in_vec		= IVLD_IDX;//we leave value unchanged for re-adding

	return element;
}




template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::insertElementsTo(const uint32_t& req, const ValueType& default_val){

	if(e_info.size() > req) return;
	e_info					.reserve(req);
	eles					.reserve(req);

	for(uint32_t i = e_info.size(); i < req; ++i){
		asrt(i == e_info.size());
		e_info	.push_back(ElementInfo(i,default_val));
		eles	.push_back(converter.fromInt(i));
		moveUp(i,eles.size()-1);
	}
	sanity(checkHeapProperty(0));
}

template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::ensureElementValsTo(const uint32_t& req, const ValueType& default_val){
	if(e_info.size() > req) return;
	e_info.resize(req,ElementInfo(IVLD_IDX,default_val));
}


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::restoreElementIf(const KeyType& element){
	const uint32_t eid = converter.toInt(element);	asrt(e_info.size() > eid);

	ElementInfo& ef 	= e_info[eid];
	if(ef.contained()) 	return;

	ef		.index_in_vec = eles.size();
	eles	.push_back(element);
	moveUp(eid,ef.index_in_vec);
	sanity(checkHeapProperty(0));
}


template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
bool Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::checkHeapProperty(uint32_t i) const{
	return i >= eles.size()
		  || ((i == 0 || betterOrEq(eles[parent(i)],eles[i])) && checkHeapProperty(leftChild(i)) && checkHeapProperty(rightChild(i)));
}



template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
template<class PV>
void Heap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::printHeap(PV& val_printer, uint32_t i,uint32_t levl,uint32_t parnt) const{
	if(i >= eles.size()) return;
	std::string in; for(uint32_t k=levl;k--;){in +=" ";}
	LOG0(in << "Heap element " << eles[i] << " with value " << val_printer(e_info[eles[i]].value) << " at levl " << levl << " position " << i << " parent " <<parnt);
	printHeap(val_printer,leftChild(i),levl+1,i);
	printHeap(val_printer, rightChild(i),levl+1,i);
}



template<class KeyType, class ValueType, class ValueCompare, class KeyTypeConverter>
bool BumpHeap<KeyType, ValueType, ValueCompare, KeyTypeConverter>::activityBump(const KeyType& element){

	typename super::ElementInfo& ef = super::e_info[super::converter.toInt(element)];
	bool ret = true;

	if((ef.value += increment) > max){
		increment	.weaken(weaken);
		int32_t i = super::e_info.size();
		while(i--){ super::e_info[i].value.weaken(weaken);}
		ret = false;
	}

	if(ef.contained()){
		super::moveUp(super::converter.toInt(element),ef.index_in_vec);}

	sanity(super::checkHeapProperty(0));
	return ret;
}






#endif /*DECAYHEAP_HPP_*/


